import imageio
import argparse
import numpy as np
import matplotlib.pyplot as plt

import torch
from easydict import EasyDict
from torchvision import transforms
from transforms import sep_transforms

from utils.flow_utils import flow_to_image, resize_flow
from utils.torch_utils import restore_model
from models.pwclite import PWCLite


class TestHelper:
    def __init__(self, cfg):
        self.cfg = EasyDict(cfg)
        self.device = (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )
        self.model = self.init_model()
        self.input_transform = transforms.Compose(
            [
                sep_transforms.Zoom(*self.cfg.test_shape),
                sep_transforms.ArrayToTensor(),
                transforms.Normalize(mean=[0, 0, 0], std=[255, 255, 255]),
            ]
        )

    def init_model(self):
        model = PWCLite(self.cfg.model)
        # print('Number fo parameters: {}'.format(model.num_parameters()))
        model = model.to(self.device)
        model = restore_model(model, self.cfg.pretrained_model)
        model.eval()
        return model

    def run(self, imgs):
        imgs = [self.input_transform(img).unsqueeze(0) for img in imgs]
        img_pair = torch.cat(imgs, 1).to(self.device)
        return self.model(img_pair)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-m", "--model", default="checkpoints/KITTI15/pwclite_ar.tar")
    parser.add_argument("-s", "--test_shape", default=[384, 640], type=int, nargs=2)
    parser.add_argument(
        "-i",
        "--img_list",
        nargs="+",
        default=["examples/img1.png", "examples/img2.png"],
    )
    args = parser.parse_args()

    cfg = {
        "model": {
            "upsample": True,
            "n_frames": len(args.img_list),
            "reduce_dense": True,
        },
        "pretrained_model": args.model,
        "test_shape": args.test_shape,
    }

    ts = TestHelper(cfg)

    imgs = [imageio.imread(img).astype(np.float32) for img in args.img_list]
    h, w = imgs[0].shape[:2]

    flow_12 = ts.run(imgs)["flows_fw"][0]

    flow_12 = resize_flow(flow_12, (h, w))
    np_flow_12 = flow_12[0].detach().cpu().numpy().transpose([1, 2, 0])

    vis_flow = flow_to_image(np_flow_12)

    fig = plt.figure()
    plt.imshow(vis_flow)
    plt.show()



# RAFT wrappers based on the official RAFT github https://github.com/princeton-vl/RAFT

from os import mkdir
from os.path import join
from os.path import isdir, dirname, abspath, exists
from inspect import getsourcefile
from urllib.request import urlretrieve

import argparse

import torch

from neurovc.contrib.raft.raft import RAFT
from neurovc.contrib.raft.utils.flow_viz import flow_to_image
from neurovc.contrib.raft.utils.utils import InputPadder

import cv2

try:
    from torch.cuda.amp import GradScaler
except (ImportError, AttributeError):
    # dummy GradScaler for PyTorch < 1.6
    class GradScaler:
        def __init__(self):
            pass

        def scale(self, loss):
            return loss

        def unscale_(self, optimizer):
            pass

        def step(self, optimizer):
            optimizer.step()

        def update(self):
            pass


try:
    from torch.utils.tensorboard import SummaryWriter
except ImportError:
    # lightweight fallback that no-ops when tensorboard is unavailable
    class SummaryWriter:
        def add_scalar(self, *args, **kwargs):
            return None

        def close(self):
            return None


class RAFTOpticalFlow:
    def __init__(
        self,
        iters=20,
        model=None,
        path=None,
        small=False,
        mixed_precision=False,
        alternate_corr=False,
        device=0,
    ):
        if model is None or not exists(model):
            model_path = join(dirname(abspath(getsourcefile(lambda: 0))), "models")
            model = join(model_path, "raft-casme2.pth")
        if not exists(model):
            if not isdir(model_path):
                mkdir(model_path)
            print("Model could not be found, downloading raft-casme2 model...")
            import ssl

            ssl._create_default_https_context = ssl._create_unverified_context
            urlretrieve(
                "https://cloud.hiz-saarland.de/s/McMNXZ5o7xteE6n/download/raft-casme2.pth",
                model,
            )
            print("done.")

        args = argparse.Namespace(
            model=model,
            path=path,
            small=small,
            mixed_precision=mixed_precision,
            alternate_corr=alternate_corr,
        )

        """ This does not work without a cuda device: """
        model = torch.nn.DataParallel(RAFT(args))
        """ And this does not work with the old state_dict: """
        # model = RAFT(args)
        model.load_state_dict(torch.load(args.model))

        self.model = model.module
        self.device = torch.device(
            f"cuda:{device}" if torch.cuda.is_available() else "cpu"
        )
        self.model.to(self.device)
        self.model.eval()
        self.last_flow = None
        self.padder = None
        self.iters = iters

    def init(self):
        self.last_flow = None
        self.padder = None

    def calc(self, ref, frame, flow=None):
        ref_torch = torch.from_numpy(ref).permute(2, 0, 1).float()
        ref_torch = ref_torch[None].to(self.device)
        frame_torch = torch.from_numpy(frame).permute(2, 0, 1).float()
        frame_torch = frame_torch[None].to(self.device)

        if self.padder is None:
            self.padder = InputPadder(ref_torch.shape)
        ref_torch, frame_torch = self.padder.pad(ref_torch, frame_torch)

        flow_low, flow_up = self.model(
            ref_torch, frame_torch, iters=self.iters, flow_init=flow, test_mode=True
        )
        flow_up = self.padder.unpad(flow_up)
        self.last_flow = flow_up[0].permute(1, 2, 0).cpu().detach().numpy()

        return self.last_flow

    def visualize(self, title="flow", flow=None):
        if flow is None and self.last_flow is not None:
            cv2.imshow(title, flow_to_image(self.last_flow))
        if flow is not None:
            cv2.imshow(title, flow_to_image(flow))


class RAFTLogger:
    SUM_FREQ = 100

    def __init__(self, model, scheduler):
        self.model = model
        self.scheduler = scheduler
        self.total_steps = 0
        self.running_loss = {}
        self.writer = None

    def _print_training_status(self):
        metrics_data = [
            self.running_loss[k] / self.SUM_FREQ
            for k in sorted(self.running_loss.keys())
        ]
        training_str = "[{:6d}, {:10.7f}] ".format(
            self.total_steps + 1, self.scheduler.get_last_lr()[0]
        )
        metrics_str = ("{:10.4f}, " * len(metrics_data)).format(*metrics_data)

        # print the training status
        print(training_str + metrics_str)

        if self.writer is None:
            self.writer = SummaryWriter()

        for k in self.running_loss:
            self.writer.add_scalar(
                k, self.running_loss[k] / self.SUM_FREQ, self.total_steps
            )
            self.running_loss[k] = 0.0

    def push(self, metrics):
        self.total_steps += 1

        for key in metrics:
            if key not in self.running_loss:
                self.running_loss[key] = 0.0

            self.running_loss[key] += metrics[key]

        if self.total_steps % self.SUM_FREQ == self.SUM_FREQ - 1:
            self._print_training_status()
            self.running_loss = {}

    def write_dict(self, results):
        if self.writer is None:
            self.writer = SummaryWriter()

        for key in results:
            self.writer.add_scalar(key, results[key], self.total_steps)

    def close(self):
        self.writer.close()



try:
    from tfan import ThermalLandmarks as ThermalLandmarks  # type: ignore
except ImportError as exc:
    raise ImportError(
        "The ThermalLandmarks class has moved to the 'tfan' package "
        "(install via 'pip install thermal-facial-alignment'). "
        f"Import error: {exc}"
    ) from exc



import torch
import numpy as np
from abc import abstractmethod
from tensorboardX import SummaryWriter
from utils.torch_utils import (
    bias_parameters,
    weight_parameters,
    load_checkpoint,
    save_checkpoint,
    AdamW,
)


class BaseTrainer:
    """
    Base class for all trainers
    """

    def __init__(
        self, train_loader, valid_loader, model, loss_func, _log, save_root, config
    ):
        self._log = _log

        self.cfg = config
        self.save_root = save_root
        self.summary_writer = SummaryWriter(str(save_root))

        self.train_loader, self.valid_loader = train_loader, valid_loader
        self.device, self.device_ids = self._prepare_device(config["n_gpu"])

        self.model = self._init_model(model)
        self.optimizer = self._create_optimizer()
        self.loss_func = loss_func

        self.best_error = np.inf
        self.i_epoch = 0
        self.i_iter = 0

    @abstractmethod
    def _run_one_epoch(self): ...

    @abstractmethod
    def _validate_with_gt(self): ...

    def train(self):
        for epoch in range(self.cfg.epoch_num):
            self._run_one_epoch()

            if self.i_epoch % self.cfg.val_epoch_size == 0:
                errors, error_names = self._validate_with_gt()
                valid_res = " ".join(
                    "{}: {:.2f}".format(*t) for t in zip(error_names, errors)
                )
                self._log.info(" * Epoch {} ".format(self.i_epoch) + valid_res)

    def _init_model(self, model):
        model = model.to(self.device)
        if self.cfg.pretrained_model:
            self._log.info(
                "=> using pre-trained weights {}.".format(self.cfg.pretrained_model)
            )
            epoch, weights = load_checkpoint(self.cfg.pretrained_model)

            from collections import OrderedDict

            new_weights = OrderedDict()
            model_keys = list(model.state_dict().keys())
            weight_keys = list(weights.keys())
            for a, b in zip(model_keys, weight_keys):
                new_weights[a] = weights[b]
            weights = new_weights
            model.load_state_dict(weights)
        else:
            self._log.info("=> Train from scratch.")
            model.init_weights()
        model = torch.nn.DataParallel(model, device_ids=self.device_ids)
        return model

    def _create_optimizer(self):
        self._log.info("=> setting Adam solver")
        param_groups = [
            {
                "params": bias_parameters(self.model.module),
                "weight_decay": self.cfg.bias_decay,
            },
            {
                "params": weight_parameters(self.model.module),
                "weight_decay": self.cfg.weight_decay,
            },
        ]

        if self.cfg.optim == "adamw":
            optimizer = AdamW(
                param_groups, self.cfg.lr, betas=(self.cfg.momentum, self.cfg.beta)
            )
        elif self.cfg.optim == "adam":
            optimizer = torch.optim.Adam(
                param_groups,
                self.cfg.lr,
                betas=(self.cfg.momentum, self.cfg.beta),
                eps=1e-7,
            )
        else:
            raise NotImplementedError(self.cfg.optim)
        return optimizer

    def _prepare_device(self, n_gpu_use):
        """
        setup GPU device if available, move model into configured device
        """
        n_gpu = torch.cuda.device_count()
        if n_gpu_use > 0 and n_gpu == 0:
            self._log.warning(
                "Warning: There's no GPU available on this machine,"
                "training will be performed on CPU."
            )
            n_gpu_use = 0
        if n_gpu_use > n_gpu:
            self._log.warning(
                "Warning: The number of GPU's configured to use is {}, "
                "but only {} are available.".format(n_gpu_use, n_gpu)
            )
            n_gpu_use = n_gpu
        device = torch.device("cuda:0" if n_gpu_use > 0 else "cpu")
        list_ids = list(range(n_gpu_use))
        return device, list_ids

    def save_model(self, error, name):
        is_best = error < self.best_error

        if is_best:
            self.best_error = error

        models = {"epoch": self.i_epoch, "state_dict": self.model.module.state_dict()}

        save_checkpoint(self.save_root, models, name, is_best)



from neurovc.contrib.flowmag.flow_models.ARFlow.models.pwclite import PWCLite


def get_model(cfg):
    if cfg.type == "pwclite":
        model = PWCLite(cfg)
    else:
        raise NotImplementedError(cfg.type)
    return model



import torch
import torch.nn.functional as F

from neurovc.contrib.flowmag.flow_models.gmflow.geometry import (
    coords_grid,
    generate_window_grid,
    normalize_coords,
)


def global_correlation_softmax(
    feature0,
    feature1,
    pred_bidir_flow=False,
):
    # global correlation
    b, c, h, w = feature0.shape
    feature0 = feature0.view(b, c, -1).permute(0, 2, 1)  # [B, H*W, C]
    feature1 = feature1.view(b, c, -1)  # [B, C, H*W]

    correlation = torch.matmul(feature0, feature1).view(b, h, w, h, w) / (
        c**0.5
    )  # [B, H, W, H, W]

    # flow from softmax
    init_grid = coords_grid(b, h, w).to(correlation.device)  # [B, 2, H, W]
    grid = init_grid.view(b, 2, -1).permute(0, 2, 1)  # [B, H*W, 2]

    correlation = correlation.view(b, h * w, h * w)  # [B, H*W, H*W]

    if pred_bidir_flow:
        correlation = torch.cat(
            (correlation, correlation.permute(0, 2, 1)), dim=0
        )  # [2*B, H*W, H*W]
        init_grid = init_grid.repeat(2, 1, 1, 1)  # [2*B, 2, H, W]
        grid = grid.repeat(2, 1, 1)  # [2*B, H*W, 2]
        b = b * 2

    prob = F.softmax(correlation, dim=-1)  # [B, H*W, H*W]

    correspondence = (
        torch.matmul(prob, grid).view(b, h, w, 2).permute(0, 3, 1, 2)
    )  # [B, 2, H, W]

    # when predicting bidirectional flow, flow is the concatenation of forward flow and backward flow
    flow = correspondence - init_grid

    return flow, prob


def local_correlation_softmax(
    feature0,
    feature1,
    local_radius,
    padding_mode="zeros",
):
    b, c, h, w = feature0.size()
    coords_init = coords_grid(b, h, w).to(feature0.device)  # [B, 2, H, W]
    coords = coords_init.view(b, 2, -1).permute(0, 2, 1)  # [B, H*W, 2]

    local_h = 2 * local_radius + 1
    local_w = 2 * local_radius + 1

    window_grid = generate_window_grid(
        -local_radius,
        local_radius,
        -local_radius,
        local_radius,
        local_h,
        local_w,
        device=feature0.device,
    )  # [2R+1, 2R+1, 2]
    window_grid = window_grid.reshape(-1, 2).repeat(b, 1, 1, 1)  # [B, 1, (2R+1)^2, 2]
    sample_coords = coords.unsqueeze(-2) + window_grid  # [B, H*W, (2R+1)^2, 2]

    sample_coords_softmax = sample_coords

    # exclude coords that are out of image space
    valid_x = (sample_coords[:, :, :, 0] >= 0) & (
        sample_coords[:, :, :, 0] < w
    )  # [B, H*W, (2R+1)^2]
    valid_y = (sample_coords[:, :, :, 1] >= 0) & (
        sample_coords[:, :, :, 1] < h
    )  # [B, H*W, (2R+1)^2]

    valid = (
        valid_x & valid_y
    )  # [B, H*W, (2R+1)^2], used to mask out invalid values when softmax

    # normalize coordinates to [-1, 1]
    sample_coords_norm = normalize_coords(sample_coords, h, w)  # [-1, 1]
    window_feature = F.grid_sample(
        feature1, sample_coords_norm, padding_mode=padding_mode, align_corners=True
    ).permute(0, 2, 1, 3)  # [B, H*W, C, (2R+1)^2]
    feature0_view = feature0.permute(0, 2, 3, 1).view(b, h * w, 1, c)  # [B, H*W, 1, C]

    corr = torch.matmul(feature0_view, window_feature).view(b, h * w, -1) / (
        c**0.5
    )  # [B, H*W, (2R+1)^2]

    # mask invalid locations
    corr[~valid] = -1e9

    prob = F.softmax(corr, -1)  # [B, H*W, (2R+1)^2]

    correspondence = (
        torch.matmul(prob.unsqueeze(-2), sample_coords_softmax)
        .squeeze(-2)
        .view(b, h, w, 2)
        .permute(0, 3, 1, 2)
    )  # [B, 2, H, W]

    flow = correspondence - coords_init
    match_prob = prob

    return flow, match_prob



import numpy as np
import scipy.signal
import aolib.util as ut
import sound
import soundrep

USE_SCIKITS_SAMPLERATE = False

if USE_SCIKITS_SAMPLERATE:
    import scikits.samplerate


def erb_filters(length, sr, nbands, low_lim, hi_lim):
    length = int(length)
    sr = float(sr)
    max_freq = sr / 2.0
    # freqs = np.fft.fftfreq(length, 1./sr)
    # freqs = freqs[freqs >= 0]
    freqs = rfftfreq(length, 1.0 / sr)
    nfreqs = len(freqs)

    if hi_lim > sr / 2:
        hi_lim = max_freq

    cos_filts = np.zeros((nfreqs, nbands))
    cutoffs = erb2freq(np.linspace(freq2erb(low_lim), freq2erb(hi_lim), 2 + nbands))
    for k in range(nbands):
        # adjacent filters overlap by 50%
        low_cutoff = cutoffs[k]
        high_cutoff = cutoffs[k + 2]
        l_ind = np.flatnonzero(freqs > low_cutoff)[0]
        h_ind = np.flatnonzero(freqs < high_cutoff)[-1]
        avg = 0.5 * (freq2erb(low_cutoff) + freq2erb(high_cutoff))
        rnge = freq2erb(high_cutoff) - freq2erb(low_cutoff)
        # Map cutoffs from -pi/2 to pi/2 interval (then to a kind of circular filter, although the value of each
        # filter component corresponds to its distance in ERB space to the average ERB value, so it's kind of nonlinear)
        cos_filts[l_ind : h_ind + 1, k] = np.cos(
            ((freq2erb(freqs[l_ind : h_ind + 1]) - avg) / rnge) * np.pi
        )

    filts = np.zeros((nfreqs, nbands + 2))
    filts[:, 1 : nbands + 1] = cos_filts
    h_ind = np.flatnonzero(freqs < cutoffs[1])[-1]
    # lowpass filter goes to the peak of first cos filter
    filts[: h_ind + 1, 0] = np.sqrt(1 - filts[: h_ind + 1, 1] ** 2)
    l_ind = np.flatnonzero(freqs > cutoffs[nbands])[0]
    filts[l_ind:, nbands + 1] = np.sqrt(1 - filts[l_ind:, nbands] ** 2)

    return filts, cutoffs, freqs


def freq2erb(freq_hz):
    return 9.265 * np.log(1 + freq_hz / (24.7 * 9.265))


def erb2freq(n_erb):
    return 24.7 * 9.265 * (np.exp(n_erb / 9.265) - 1)


# consistent with matlab code
def subbands_from_sound(signal, filts):
    ft_signal = np.fft.rfft(signal)
    ft_subbands = filts * ft_signal[:, np.newaxis]
    subbands = np.fft.irfft(ft_subbands, axis=0)
    # ut.toplevel_locals()
    return subbands.real
    # if 0:
    #   fft_signal = np.fft.fft(signal)
    #   fft_filts = np.vstack([filts, filts[1:-1][::-1]])
    #   fft_subbands = fft_filts * fft_signal[:, np.newaxis]
    #   subbands = np.real(np.fft.ifft(fft_subbands, axis = 0))
    #   ut.toplevel_locals()
    #   return subbands
    # #ft_signal = np.fft.fft(signal)


def sound_from_subbands(subbands, filts):
    ft_subbands = np.fft.rfft(subbands, axis=0)
    ft = filts * ft_subbands
    ift = np.fft.irfft(ft, axis=0)
    return np.sum(ift, axis=1)


def resample(signal, sc, clip=True, num_samples=None):
    # print 'Called', clip
    # ut.tic()
    if USE_SCIKITS_SAMPLERATE:
        r = np.array(scikits.samplerate.resample(signal, sc, "sinc_best"), "d")
        assert not np.any(np.isinf(r)) and not np.any(np.isnan(r))
    else:
        # print signal.shape[0], int(round(signal.shape[0] * sc))
        n = int(round(signal.shape[0] * sc)) if num_samples is None else num_samples
        # print 'converting:', signal.shape[0], '->', n
        r = scipy.signal.resample(signal, n)

    if clip:
        r = np.clip(r, -1, 1)
    # ut.toc()
    return r


def resample_sound(signal, sc, clip=True, num_samples=None):
    n = int(round(signal.shape[0] * sc)) if num_samples is None else num_samples
    print("resampling", signal.shape[0], n)
    r = scipy.signal.resample(signal, n)
    if clip:
        r = np.clip(r, -1, 1)
    return r


def cut_signal(signal, new_sr, old_sr):
    new_signal = signal[: good_resample_length(signal.shape[0], new_sr, old_sr)]
    # print signal.shape[0], new_signal.shape[0]
    return new_signal

    # r = max(1, np.floor(old_sr / new_sr)) * (new_sr / 100)
    # r = 100
    # return signal[:int(np.floor(signal.shape[0] / r) * r)]


# original
# def good_resample_length(n, new_sr, old_sr, factor = 100):
#   assert new_sr % factor == 0 and old_sr % factor == 0
#   return int(np.floor(n / factor) * factor)

# def good_resample_length(n, new_sr, old_sr, factor = 100):
#   factor = new_sr / float(old_sr)
#   new_len = int(np.floor(n / factor) * factor)
#   print 'New length:', new_len, 'Factor:', factor
#   return new_len


def good_resample_length(n, new_sr, old_sr, factor2=100):
    factor = new_sr / float(old_sr)
    new_len = int(np.floor(n / factor) * factor)
    return ut.make_mod(new_len, factor2)


# def good_resample_length(n, new_sr, old_sr, factor2 = 100):
#   factor = new_sr / float(old_sr)
#   for i in xrange(1+factor2):
#     nn = (n - i)
#     #new_len = int(np.floor(nn / factor) * factor)
#     new_len = int(round(nn * factor))
#     print nn, new_len
#     if new_len % factor2 == 0:
#       return nn
#   return ut.make_mod(n, factor2)

# def rescale_waveform_rms(snd, desired_rms):
#   signal0 = snd.normalized().to_mono().samples.flatten()
#   est_rms = max(1e-5, np.mean(signal0**2))
#   signal0 = signal0 * (desired_rms / est_rms)
#   return signal0


# def subband_envs(snd, nbands, env_sr, num_samples = None,
#                  ds_sr = 20000, low_lim = 20, hi_lim = 10000, comp_exp = 0.3, desired_rms = None, clip_samples = True):
#   ds_sr = float(ds_sr)
#   signal0, sr0 = snd.normalized().to_mono().samples.flatten(), snd.rate
#   if desired_rms is not None:
#     #eps = np.finfo(np.float64).eps
#     #print signal0.max(), desired_rms, np.mean(signal0**2)
#     #signal0 = signal0 * (desired_rms / (eps + np.mean(signal0**2)))
#     est_rms = max(1e-5, np.mean(signal0**2))
#     signal0 = signal0 * (desired_rms / est_rms)
#     #print signal0.max()

#   sr0 = float(sr0)
#   signal0 = cut_signal(signal0, ds_sr, sr0)
#   signal = resample(signal0, ds_sr / sr0, clip = clip_samples)

#   # number of samples in the signal that correspond to one subband envelope feature
#   ds_factor = int(np.floor(ds_sr / env_sr))

#   signal = signal[:int((len(signal)//ds_factor) * ds_factor)]
#   audio_filts, Hz_cutoffs, freqs = erb_filters(len(signal), ds_sr, nbands, low_lim, hi_lim) # consistent w/ matlab
#   subbands = subbands_from_sound(signal, audio_filts)
#   analytic_subbands = scipy.signal.hilbert(subbands, axis = 0)
#   subband_envs = np.abs(analytic_subbands)
#   subband_envs **= comp_exp
#   subband_envs = resample(subband_envs, env_sr / ds_sr, num_samples = num_samples, clip = clip_samples)

#   return subband_envs


def subband_envs(
    snd,
    nbands,
    env_sr,
    num_samples=None,
    ds_sr=20000,
    low_lim=20,
    hi_lim=10000,
    comp_exp=0.3,
    desired_rms=None,
    clip_samples=True,
):
    ds_sr = float(ds_sr)
    signal0, sr0 = snd.normalized().to_mono().samples.flatten(), snd.rate
    if desired_rms is not None:
        # est_rms = max(1e-5, np.mean(signal0**2))
        est_rms = max(1e-5, ut.rms(signal0))
        signal0 = signal0 * (desired_rms / est_rms)
        # print 'rms', ut.rms(signal0)

    sr0 = float(sr0)
    signal0 = cut_signal(signal0, ds_sr, sr0)
    signal = resample(signal0, ds_sr / sr0, clip=clip_samples)

    # number of samples in the signal that correspond to one subband envelope feature
    ds_factor = int(np.floor(ds_sr / env_sr))

    signal = signal[: int((len(signal) // ds_factor) * ds_factor)]
    audio_filts, Hz_cutoffs, freqs = erb_filters(
        len(signal), ds_sr, nbands, low_lim, hi_lim
    )  # consistent w/ matlab
    subbands = subbands_from_sound(signal, audio_filts)
    analytic_subbands = scipy.signal.hilbert(subbands, axis=0)
    subband_envs = np.abs(analytic_subbands)
    subband_envs **= comp_exp
    subband_envs = resample(
        subband_envs, env_sr / ds_sr, num_samples=num_samples, clip=clip_samples
    )

    return subband_envs


def invert_subband_envs(
    target_envs,
    final_sr,
    env_sr,
    mid_sr=20000,
    niters=3,
    low_lim=20,
    hi_lim=10000,
    comp_exp=0.3,
):
    # number of subbands (w/o high/low pass)
    nbands = target_envs.shape[1] - 2
    snd_len = int(np.ceil(target_envs.shape[0] * float(mid_sr) / float(env_sr)))
    snd_len -= snd_len % 100
    target_envs = np.clip(target_envs, 0, 1)

    # snd_len = good_resample_length(target_envs.shape[0], mid_sr, env_sr, 10)
    filts, _, _ = erb_filters(snd_len, mid_sr, nbands, low_lim, hi_lim)
    # cut_signal(signal, new_sr, old_sr):
    synth_sound = np.random.randn(snd_len)
    for i in range(niters):
        # Forward pass: current sound -> downsampled envelopes and full-res phases
        synth_subbands = subbands_from_sound(synth_sound, filts)
        analytic = scipy.signal.hilbert(synth_subbands, axis=0)
        synth_envs = np.abs(analytic)
        phases = analytic / synth_envs

        # up_target_envs = resample(target_envs, mid_sr / env_sr)
        up_target_envs = scipy.signal.resample(target_envs, phases.shape[0])
        up_target_envs = np.maximum(up_target_envs, 0.0)
        up_target_envs **= 1.0 / comp_exp

        new_analytic = phases * up_target_envs
        synth_subbands = np.real(new_analytic)
        synth_sound = sound_from_subbands(synth_subbands, filts)

    synth_sound = resample(synth_sound, final_sr / float(mid_sr))
    synth_sound = np.clip(synth_sound, -1.0, 1.0)
    return sound.Sound(None, final_sr, synth_sound)


def invert_noisy_subbands(
    target_envs,
    final_sr,
    env_sr,
    mid_sr=20000,
    low_lim=20,
    hi_lim=10000,
    comp_exp=0.3,
    scale_volume=2.0,
):
    # number of subbands (w/o high/low pass)
    nbands = target_envs.shape[1] - 2
    snd_len = int(np.ceil(target_envs.shape[0] * float(mid_sr) / float(env_sr)))
    snd_len -= snd_len % 100
    target_envs = np.clip(target_envs, 0, 1)

    filts, _, _ = erb_filters(snd_len, mid_sr, nbands, low_lim, hi_lim)
    synth_sound = np.random.randn(snd_len)
    if 0:
        print("using pink noise")
        noise_snd = sound.load_sound(
            "/data/vision/billf/aho-stuff/vis/lib/soundtex/pink_noise_20s_20kHz.wav"
        )
        noise_snd = noise_snd.normalized().to_mono()
        noise_snd = noise_snd.samples
        i = np.random.choice(list(range(noise_snd.shape[0] - synth_sound.shape[0])))
        synth_sound = 100 * noise_snd[i : i + synth_sound.shape[0]]
        print("stdev", np.std(synth_sound))
        # sound.play(synth_sound, mid_sr)

    # Forward pass: current sound -> downsampled envelopes and full-res phases
    synth_subbands = subbands_from_sound(synth_sound, filts)
    analytic = scipy.signal.hilbert(synth_subbands, axis=0)
    phases = analytic

    # up_target_envs = resample(target_envs, mid_sr / env_sr)
    up_target_envs = scipy.signal.resample(target_envs, phases.shape[0])
    up_target_envs = np.maximum(up_target_envs, 0.0)
    up_target_envs **= 1.0 / comp_exp

    new_analytic = phases * up_target_envs
    synth_subbands = np.real(new_analytic)
    synth_sound = sound_from_subbands(synth_subbands, filts)

    synth_sound = resample(synth_sound, final_sr / float(mid_sr))
    synth_sound = synth_sound * scale_volume
    synth_sound = np.clip(synth_sound, -1.0, 1.0)
    return sound.Sound(None, final_sr, synth_sound)


def reencode_subbands(
    snd, nbands, env_sr, matlab_encode=False, matlab_decode=False, do_decode=True
):
    if matlab_encode:
        envs = soundrep.matlab_subband(snd, env_sr)
    else:
        envs = subband_envs(snd, nbands, env_sr)

    if do_decode:
        ut.toplevel_locals()
        if matlab_decode:
            return sound.Sound(
                None,
                snd.rate,
                soundrep.matlab_inv_subband(envs, snd.rate, env_sr, nbands=nbands),
            )
        return invert_subband_envs(envs, snd.rate, env_sr)


def rfftfreq(n, d=1.0):
    if not (isinstance(n, int) or isinstance(n, np.integer)):
        raise ValueError("n should be an integer")
    val = 1.0 / (n * d)
    N = n // 2 + 1
    results = np.arange(0, N, dtype=np.int)
    return results * val


# def mod_filters(length, sr, num_mod_channels = 20, low_lim = 0.5, hi_lim = 200., circular_conv = True, q = 2.):
def mod_filters(
    length,
    sr,
    num_mod_channels=10,
    low_lim=0.5,
    hi_lim=200.0,
    circular_conv=True,
    q=1.0,
):
    """Modulation filters. These are similar to the cochlear filters
    (make_erb_cos_filters), but they're applied to subband envs, rather
    than to the raw waveform, and so they are tuned to lower
    frequencies."""
    # see make_constQ_cos_filters.m
    max_freq = sr / 2.0
    freqs = rfftfreq(length, 1.0 / sr)
    nfreqs = len(freqs)
    if hi_lim > sr / 2:
        hi_lim = max_freq

    # center frequencies evenly spaced on log scale
    cos_filts = np.zeros((nfreqs, num_mod_channels))
    cfs = 2 ** np.linspace(np.log2(low_lim), np.log2(hi_lim), num_mod_channels)
    for k in range(num_mod_channels):
        # increase the width of the filter as a function of frequency
        bw = cfs[k] / q
        low_bound = cfs[k] - bw
        high_bound = cfs[k] + bw
        freq_ok = np.flatnonzero((low_bound < freqs) & (freqs < high_bound))
        l_ind, h_ind = freq_ok[0], 1 + freq_ok[-1]
        avg = cfs[k]
        rnge = high_bound - low_bound
        # go from theta = -pi/2 to pi/2
        filt = np.cos(((freqs[l_ind:h_ind] - avg) / rnge) * np.pi)
        cos_filts[l_ind:h_ind, k] = filt

    # the filters won't sum to 1 (not sure why this only applies to the middle of the spectrum)
    # this probably doesn't matter as much for us, since we're not synthesizing sound
    energy = np.sum(cos_filts**2, axis=1)
    energy_middle = energy[(freqs >= cfs[3]) & (freqs <= cfs[-4])]
    filts = cos_filts / np.sqrt(energy_middle.mean())
    return filts


def mod_power(envs, sr):
    filts = mod_filters(envs.shape[0], sr)
    mod_subbands = []
    for j in range(envs.shape[1]):
        mod_subbands.append(subbands_from_sound(envs[:, j], filts))
    return np.array(mod_subbands).transpose((1, 0, 2))


# def texture_stats(envs, sr, normalize_envs = False):
#   # marginal subband stats
#   # not using a windowing function
#   loudness = np.median(np.sqrt(np.sum(envs**2, axis = 1)))
#   if normalize_envs:
#     envs = envs / loudness

#   env_mu = envs.mean(0)
#   env_var_unnorm = envs.var(0)
#   env_var_norm = env_var_unnorm / env_mu**2
#   env_stdev_norm = np.sqrt(env_var_norm)

#   mods = mod_power(envs, sr)
#   mod_pow_norm = np.zeros(mods.shape[1:])
#   for j in xrange(envs.shape[1]):
#     mod_pow_unnorm = np.mean(mods[:, j]**2, axis = 0)
#     mod_pow_norm[j, :] = np.sqrt(mod_pow_unnorm / env_var_unnorm[j])

#   return env_mu, env_stdev_norm

# not used in final version
# def texture_stats(envs, sr, normalize_envs = False):
#   # marginal subband stats
#   # not using a windowing function
#   envs = envs.astype('float64')
#   if normalize_envs:
#     loudness = np.median(np.sqrt(np.sum(envs**2, axis = 1)))
#     envs = envs / loudness
#     assert 0
#   else:
#     loudness = np.array(0.)

#   eps = np.finfo(np.float32).eps
#   env_mu = envs.mean(0)
#   env_var_unnorm = envs.var(0)
#   env_var_norm = env_var_unnorm / (eps + env_mu**2)
#   env_stdev_norm = np.sqrt(env_var_norm)

#   pairs = [1, 2, 3, 5]
#   num_bands = envs.shape[1]
#   corrs = []
#   for i in xrange(num_bands):
#     for j in pairs:
#       ii = i + j
#       if ii < num_bands:
#         corrs.append(np.corrcoef(envs[:, i], envs[:, ii])[0, 1])
#   corrs = np.array(corrs)
#   # when a band has zero variance, this can happen
#   corrs[np.isnan(corrs)] = 0.

#   mods = mod_power(envs, sr)
#   mod_pow_norm = np.zeros(mods.shape[1:])
#   for j in xrange(num_bands):
#     mod_pow_unnorm = np.mean(mods[:, j]**2, axis = 0)
#     mod_pow_norm[j, :] = np.sqrt(mod_pow_unnorm / (eps + env_var_unnorm[j]))

#   for x in [env_mu, env_stdev_norm, corrs, mod_pow_norm, loudness]:
#     assert not np.any(np.isnan(x))
#     assert not np.any(np.isinf(x))

#   return env_mu, env_stdev_norm, corrs, mod_pow_norm, loudness


def texture_stats(envs, sr, normalize_envs=False):
    # marginal subband stats
    # not using a windowing function
    envs = envs.astype("float64")
    if normalize_envs:
        loudness = np.median(np.sqrt(np.sum(envs**2, axis=1)))
        envs = envs / loudness
        # assert 0
    else:
        loudness = np.array(0.0)

    eps = np.finfo(np.float32).eps
    env_mu = envs.mean(0)
    env_var_unnorm = envs.var(0)
    env_var_norm = env_var_unnorm / (eps + env_mu**2)
    env_stdev_norm = np.sqrt(env_var_norm)

    pairs = [1, 2, 3, 5]
    num_bands = envs.shape[1]
    corrs = []
    for i in range(num_bands):
        for j in pairs:
            ii = i + j
            if ii < num_bands:
                corrs.append(np.corrcoef(envs[:, i], envs[:, ii])[0, 1])
    corrs = np.array(corrs)
    # when a band has zero variance, this can happen
    corrs[np.isnan(corrs)] = 0.0

    mods = mod_power(envs, sr)
    mod_pow_norm = np.zeros(mods.shape[1:])
    for j in range(num_bands):
        mod_pow_unnorm = np.mean(mods[:, j] ** 2, axis=0)
        mod_pow_norm[j, :] = np.sqrt(mod_pow_unnorm / (eps + env_var_unnorm[j]))

    for x in [env_mu, env_stdev_norm, corrs, mod_pow_norm, loudness]:
        assert not np.any(np.isnan(x))
        assert not np.any(np.isinf(x))

    return env_mu, env_stdev_norm, corrs, mod_pow_norm, loudness


def normalize_subbands(envs):
    loudness = np.median(np.sqrt(np.sum(envs**2, axis=1)))
    envs = envs / (0.01 + loudness)
    return envs



import torch.nn as nn
import torch.nn.functional as F
from neurovc.contrib.flowmag.flow_models.ARFlow.losses.loss_blocks import (
    SSIM,
    TernaryLoss,
    smooth_grad_1st,
    smooth_grad_2nd,
)
from utils.warp_utils import flow_warp
from utils.warp_utils import get_occu_mask_bidirection, get_occu_mask_backward


class unFlowLoss(nn.modules.Module):
    def __init__(self, cfg):
        super(unFlowLoss, self).__init__()
        self.cfg = cfg

    def loss_photomatric(self, im1_scaled, im1_recons, occu_mask1):
        loss = []

        if self.cfg.w_l1 > 0:
            loss += [self.cfg.w_l1 * (im1_scaled - im1_recons).abs() * occu_mask1]

        if self.cfg.w_ssim > 0:
            loss += [
                self.cfg.w_ssim * SSIM(im1_recons * occu_mask1, im1_scaled * occu_mask1)
            ]

        if self.cfg.w_ternary > 0:
            loss += [
                self.cfg.w_ternary
                * TernaryLoss(im1_recons * occu_mask1, im1_scaled * occu_mask1)
            ]

        return sum(item.mean() for item in loss) / occu_mask1.mean()

    def loss_smooth(self, flow, im1_scaled):
        if "smooth_2nd" in self.cfg and self.cfg.smooth_2nd:
            func_smooth = smooth_grad_2nd
        else:
            func_smooth = smooth_grad_1st
        loss = []
        loss += [func_smooth(flow, im1_scaled, self.cfg.alpha)]
        return sum(item.mean() for item in loss)

    def forward(self, output, target):
        """

        :param output: Multi-scale forward/backward flows n * [B x 4 x h x w]
        :param target: image pairs Nx6xHxW
        :return:
        """

        pyramid_flows = output
        im1_origin = target[:, :3]
        im2_origin = target[:, 3:]

        pyramid_smooth_losses = []
        pyramid_warp_losses = []
        self.pyramid_occu_mask1 = []
        self.pyramid_occu_mask2 = []

        s = 1.0
        for i, flow in enumerate(pyramid_flows):
            if self.cfg.w_scales[i] == 0:
                pyramid_warp_losses.append(0)
                pyramid_smooth_losses.append(0)
                continue

            b, _, h, w = flow.size()

            # resize images to match the size of layer
            im1_scaled = F.interpolate(im1_origin, (h, w), mode="area")
            im2_scaled = F.interpolate(im2_origin, (h, w), mode="area")

            im1_recons = flow_warp(im2_scaled, flow[:, :2], pad=self.cfg.warp_pad)
            im2_recons = flow_warp(im1_scaled, flow[:, 2:], pad=self.cfg.warp_pad)

            if i == 0:
                if self.cfg.occ_from_back:
                    occu_mask1 = 1 - get_occu_mask_backward(flow[:, 2:], th=0.2)
                    occu_mask2 = 1 - get_occu_mask_backward(flow[:, :2], th=0.2)
                else:
                    occu_mask1 = 1 - get_occu_mask_bidirection(flow[:, :2], flow[:, 2:])
                    occu_mask2 = 1 - get_occu_mask_bidirection(flow[:, 2:], flow[:, :2])
            else:
                occu_mask1 = F.interpolate(
                    self.pyramid_occu_mask1[0], (h, w), mode="nearest"
                )
                occu_mask2 = F.interpolate(
                    self.pyramid_occu_mask2[0], (h, w), mode="nearest"
                )

            self.pyramid_occu_mask1.append(occu_mask1)
            self.pyramid_occu_mask2.append(occu_mask2)

            loss_warp = self.loss_photomatric(im1_scaled, im1_recons, occu_mask1)

            if i == 0:
                s = min(h, w)

            loss_smooth = self.loss_smooth(flow[:, :2] / s, im1_scaled)

            if self.cfg.with_bk:
                loss_warp += self.loss_photomatric(im2_scaled, im2_recons, occu_mask2)
                loss_smooth += self.loss_smooth(flow[:, 2:] / s, im2_scaled)

                loss_warp /= 2.0
                loss_smooth /= 2.0

            pyramid_warp_losses.append(loss_warp)
            pyramid_smooth_losses.append(loss_smooth)

        pyramid_warp_losses = [
            loss_value * weight
            for loss_value, weight in zip(pyramid_warp_losses, self.cfg.w_scales)
        ]
        pyramid_smooth_losses = [
            loss_value * weight
            for loss_value, weight in zip(pyramid_smooth_losses, self.cfg.w_sm_scales)
        ]

        warp_loss = sum(pyramid_warp_losses)
        smooth_loss = self.cfg.w_smooth * sum(pyramid_smooth_losses)
        total_loss = warp_loss + smooth_loss

        return total_loss, warp_loss, smooth_loss, pyramid_flows[0].abs().mean()



import matplotlib.pyplot as plt
import torch
from torch.optim import Adam
from torch.utils.data import DataLoader
from tqdm import tqdm

from neurovc.contrib.flowmag.dataset import TestTimeAdaptDataset
from neurovc.contrib.flowmag.myutils import AverageMeter


def test_time_adapt(
    model,
    frames_dir,
    num_epochs=5,
    mode="first",
    device=0,
    inference_fn=None,
    inference_freq=1,
    alpha=None,
    save_dir=None,
    dataset_length=None,
):
    """
    params:
        model: (nn.Module) model with checkpoint already loaded
        frames_dir: (string) path to directory of frames for test time adaptation (OPTIONAL NOW)
        dataset: optional dataset to give, default is TestTimeAdaptDataset
        num_epochs: (int) number of passes through the frames_dir
        mode: ['first', 'random'] how to sample frames
        device: device to put model and data on
        inference_fn: function to call at the end of each epoch

    output:
        model: (nn.Module) finetuned input module
    """

    model.train()
    model = model.to(device)

    # Get dataset from frames
    # if dataset is None:
    dataset = TestTimeAdaptDataset(frames_dir, mode=mode, length=dataset_length)
    dataloader = DataLoader(
        dataset, batch_size=8, shuffle=True, num_workers=4, drop_last=False
    )

    # Optimizer
    optimizer = Adam(model.module.trainable_parameters(), lr=1e-4)

    # Record losses
    meter_loss = AverageMeter("loss")
    meter_mag_loss = AverageMeter("loss_mag")
    meter_color_loss = AverageMeter("loss_color")

    # Record average of losses
    hist_loss = []
    hist_mag_loss = []
    hist_color_loss = []

    for epoch in range(num_epochs):
        print(f"Epoch {epoch}/{num_epochs}")

        # Run inference
        if inference_fn is not None and epoch % inference_freq == 0:
            print("Performing inference...")
            model.eval()
            inference_fn(model, epoch)
            model.train()

        # Save checkpoints of tta
        if save_dir is not None:
            print("saving epoch checkpoint...")
            chkpt = {
                "epoch": epoch,
                "state_dict": model.state_dict(),
                "optimizer": optimizer.state_dict(),
            }
            torch.save(chkpt, save_dir / f"chkpt_{epoch:04}.pth")

        for cur_iter, frames in enumerate(tqdm(dataloader)):
            # Process data
            frames = frames.to(device)

            # Model
            preds, loss, info = model(frames, alpha=alpha)
            loss = loss.mean()

            # Optimize
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            # Track losses
            meter_loss.update(loss.item())
            meter_mag_loss.update(info["loss_mag"].item())
            meter_color_loss.update(info["loss_color"].item())

        print(f"Avg Loss: {meter_loss.avg}")
        print(f"Avg Mag Loss: {meter_mag_loss.avg}")
        print(f"Avg Color Loss: {meter_color_loss.avg}")
        hist_loss.append(meter_loss.avg)
        hist_mag_loss.append(meter_mag_loss.avg)
        hist_color_loss.append(meter_color_loss.avg)
        meter_loss.reset()
        meter_mag_loss.reset()
        meter_color_loss.reset()

        # Plot losses
        if save_dir is not None:
            loss_info = {
                "losses": hist_loss,
                "mag_losses": hist_mag_loss,
                "color_losses": hist_color_loss,
            }
            for loss_name, losses in loss_info.items():
                plt.plot(losses)
                plt.title(loss_name)
                plt.savefig(save_dir / f"{loss_name}.png")
                plt.clf()

    # Final inference
    model.eval()
    if inference_fn is not None:
        inference_fn(model, epoch)

    info = {
        "losses": hist_loss,
        "mag_losses": hist_mag_loss,
        "color_losses": hist_color_loss,
    }

    return model, info



import os
import dominate
from dominate.tags import a, br, h3, img, meta, p, table, td, tr

# create a html file to display the gif files of frame pairs


class HTML:
    def __init__(self, web_dir, title, refresh=0):
        self.title = title
        self.web_dir = web_dir
        self.img_dir = os.path.join(self.web_dir, "images")
        if not os.path.exists(self.web_dir):
            os.makedirs(self.web_dir)
        if not os.path.exists(self.img_dir):
            os.makedirs(self.img_dir)

        self.doc = dominate.document(title=title)
        if refresh > 0:
            with self.doc.head:
                meta(http_equiv="refresh", content=str(refresh))

    def get_image_dir(self):
        return self.img_dir

    def add_header(self, str):
        with self.doc:
            h3(str)

    def add_table(self, border=1):
        self.t = table(border=border, style="table-layout: fixed;")
        self.doc.add(self.t)

    def add_images(self, ims, txts, links, width=256):
        self.add_table()
        with self.t:
            with tr():
                for im, txt, link in zip(ims, txts, links):
                    with td(
                        style="word-wrap: break-word;", halign="center", valign="top"
                    ):
                        with p():
                            with a(href=os.path.join("images", link)):
                                img(
                                    style="width:%dpx" % (width),
                                    src=os.path.join("images", im),
                                )
                            br()
                            p(txt)

    def save(self):
        html_file = "%s/index.html" % self.web_dir
        f = open(html_file, "wt")
        f.write(self.doc.render())
        f.close()


def save_gif(image_paths, save_root):
    import imageio.v2 as imageio

    save_fn = "{}_{}.gif".format(
        image_paths[0].split("/")[-2], os.path.basename(image_paths[0]).split(".")[0]
    )

    with imageio.get_writer(os.path.join(save_root, save_fn), mode="I") as writer:
        for image_path in image_paths:
            image = imageio.imread(image_path)
            writer.append_data(image)
    return save_fn
